package util;

import java.math.BigInteger;
import java.security.GeneralSecurityException;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.Signature;
import java.security.SignatureException;
import java.security.interfaces.DSAParams;
import java.security.interfaces.DSAPrivateKey;
import java.security.interfaces.DSAPublicKey;
import java.util.Arrays;

import org.apache.commons.codec.binary.Base64;

public class XmlDigitalSigner
{
    private static String base64(byte[] bytes)
    {
        return new Base64(Integer.MAX_VALUE).encodeToString(bytes).trim();
    }
    
    private static String base64(BigInteger n)
    {
        byte[] bytes = n.toByteArray();
        
        if (bytes[0] == 0) {
            bytes = Arrays.copyOfRange(bytes, 1, bytes.length);
        }
        
        return base64(bytes);
    }
    
    private static byte[] sha1(byte[] source) throws NoSuchAlgorithmException
    {
        if (source == null) {
            return null;
        }
        
        return MessageDigest.getInstance("SHA-1").digest(source);
    }
    
    private static String sha1(String source) throws NoSuchAlgorithmException
    {
        return base64(sha1(source.getBytes()));
    }
    
    private static byte[] fit(byte[] b, int length)
    {
        int len = b.length;
        int offset = (b[0] == 0) ? 1 : 0;
        len -= offset;
        
        if (len > length) {
           throw new IllegalArgumentException();
        }
        
        byte[] bytes = new byte[length];
        
        System.arraycopy(b, offset, bytes, length - len, len);
        
        return bytes;
    }
    
    private static byte[] decodeSignature(byte[] signature)
    {
        int start = 4;
        int len = signature[start - 1];
        
        byte[] data = new byte[len];
        System.arraycopy(signature, start, data, 0, len);
        byte[] r = fit(data, 20);

        start += len + 2;
        len = signature[start - 1];
        data = new byte[len];
        System.arraycopy(signature, start, data, 0, len);
        byte[] s = fit(data, 20);

        byte[] b = new byte[40];
        System.arraycopy(r, 0, b,  0, 20);
        System.arraycopy(s, 0, b, 20, 20);

        return b;
    }
    
    private static String sign(String signedInfo, DSAPrivateKey key) throws NoSuchAlgorithmException, InvalidKeyException, SignatureException
    {
        Signature sign = Signature.getInstance("SHA1withDSA");
        sign.initSign(key);
        sign.update(signedInfo.getBytes());
        
        return base64(decodeSignature(sign.sign()));
    }
    
    public static String signXml(String xml, DSAPublicKey publicKey, DSAPrivateKey privateKey) throws SamlException
    {
        try {
            String digest = sha1(xml.replace(SIGNATURE_PLACEHOLDER, ""));
            String signedInfo = SIGNED_INFO_TEMPLATE.replace("${DIGEST_VALUE}", digest);
            
            String signature = SIGNATURE_TEMPLATE;
            signature = signature.replace("${SIGNED_INFO}", signedInfo);
            signature = signature.replace("${SIGNATURE_VALUE}", sign(signedInfo, privateKey));
            
            DSAParams params = publicKey.getParams();
            signature = signature.replace("${P}", base64(params.getP()));
            signature = signature.replace("${Q}", base64(params.getQ()));
            signature = signature.replace("${G}", base64(params.getG()));
            signature = signature.replace("${Y}", base64(publicKey.getY()));
            
            return xml.replace(SIGNATURE_PLACEHOLDER, signature);
        }
        catch (GeneralSecurityException e) {
            throw new SamlException("Error signing SAML element: " + e.getMessage());
        }
    }
    
    private static final String SIGNED_INFO_TEMPLATE = "<SignedInfo xmlns=\"http://www.w3.org/2000/09/xmldsig#\" xmlns:samlp=\"urn:oasis:names:tc:SAML:2.0:protocol\" xmlns:xenc=\"http://www.w3.org/2001/04/xmlenc#\"><CanonicalizationMethod Algorithm=\"http://www.w3.org/TR/2001/REC-xml-c14n-20010315#WithComments\"></CanonicalizationMethod><SignatureMethod Algorithm=\"http://www.w3.org/2000/09/xmldsig#dsa-sha1\"></SignatureMethod><Reference URI=\"\"><Transforms><Transform Algorithm=\"http://www.w3.org/2000/09/xmldsig#enveloped-signature\"></Transform></Transforms><DigestMethod Algorithm=\"http://www.w3.org/2000/09/xmldsig#sha1\"></DigestMethod><DigestValue>${DIGEST_VALUE}</DigestValue></Reference></SignedInfo>";
    
    private static final String SIGNATURE_TEMPLATE = "<Signature xmlns=\"http://www.w3.org/2000/09/xmldsig#\">${SIGNED_INFO}<SignatureValue>${SIGNATURE_VALUE}</SignatureValue><KeyInfo><KeyValue><DSAKeyValue><P>${P}</P><Q>${Q}</Q><G>${G}</G><Y>${Y}</Y></DSAKeyValue></KeyValue></KeyInfo></Signature>";
    
    private static final String SIGNATURE_PLACEHOLDER = "${SIGNATURE}";
}
